"""gRPC non-blocking calls utils."""
import logging
from typing import TYPE_CHECKING, Any, Callable

import grpc
from google.protobuf.empty_pb2 import Empty

if TYPE_CHECKING:
    from .client import GrpcClient

_LOGGER = logging.getLogger(__name__)


class FutureCallWrapper:
    """gRPC future call wrapper."""

    def __init__(
        self,
        future_call: grpc.Future,
        *,
        client: 'GrpcClient',
        service: str,
        method: str,
        call_id: str,
        modify_result: Callable[[Any], Any] = None
    ):
        """Initialize future call wrapper."""
        self._future_call = future_call
        self._client = client
        self._service = service
        self._method = method
        self._call_id = call_id
        if modify_result is None:
            self._modify_result = lambda x: x
        else:
            self._modify_result = modify_result

    def result(self, timeout=None) -> Any:
        """Decode the result and possibly throw an exception."""
        try:
            try:
                response = self._future_call.result(timeout)
            except grpc.RpcError as error:
                if error.code() == grpc.StatusCode.UNAVAILABLE:
                    # Connection probably failed. Try to reconnect and send the request again.
                    _LOGGER.info('Connection to gRPC service %s failed. Trying to reconnect.', self._service)
                    self._client.connect()
                    response = self._future_call.result(timeout)
                else:
                    raise
        except grpc.RpcError as error:
            _LOGGER.info('[%s] %s failed with %s', self._call_id, self._method, error)
            raise

        _LOGGER.debug('[%s] %s returned %s', self._call_id, self._method, response)

        # Assert the supported structure of messages
        assert ('exception' in response.DESCRIPTOR.fields_by_name or 'data' in response.DESCRIPTOR.fields_by_name
                or isinstance(response, Empty))

        # Decode exception first
        if 'exception' in response.DESCRIPTOR.fields_by_name and response.HasField('exception'):
            raise self._client.decoder.decode_exception(response.exception)  # type: ignore

        # Decode data otherwise
        if 'data' in response.DESCRIPTOR.fields_by_name and response.HasField('data'):
            return self._modify_result(self._client.decoder.decode(response.data))  # type: ignore

        # The response is empty
        return self._modify_result(None)

    def __getattr__(self, name):
        """Proxy all other future properties."""
        return getattr(self._future_call, name)
